"use strict";
/**
 * @copyright Cube Dev, Inc.
 * @license Apache-2.0
 * @fileoverview The `MongoBIDriver` and related types declaration.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.MongoBIDriver = void 0;
const shared_1 = require("@cubejs-backend/shared");
const mysql2_1 = require("mysql2");
const generic_pool_1 = __importDefault(require("generic-pool"));
const base_driver_1 = require("@cubejs-backend/base-driver");
const MySQLType_1 = require("./MySQLType");
/**
 * MongoBI driver class.
 */
class MongoBIDriver extends base_driver_1.BaseDriver {
    /**
     * Returns default concurrency value.
     */
    static getDefaultConcurrency() {
        return 2;
    }
    config;
    pool;
    /**
     * Class constructor.
     */
    constructor(config = {}) {
        super({
            testConnectionTimeout: config.testConnectionTimeout,
        });
        // eslint-disable-next-line @typescript-eslint/no-unused-vars
        const { dataSource: configDataSource, maxPoolSize, testConnectionTimeout, ...mongoBIDriverConfiguration } = config;
        const dataSource = configDataSource || (0, shared_1.assertDataSource)('default');
        this.config = {
            host: (0, shared_1.getEnv)('dbHost', { dataSource }),
            database: (0, shared_1.getEnv)('dbName', { dataSource }),
            port: (0, shared_1.getEnv)('dbPort', { dataSource }),
            user: (0, shared_1.getEnv)('dbUser', { dataSource }),
            password: (0, shared_1.getEnv)('dbPass', { dataSource }),
            // mysql2 uses own typings for ssl property, which is not correct
            // Types of property 'pfx' are incompatible. Skipping validation with any cast
            ssl: this.getSslOptions(dataSource),
            typeCast: (field, next) => {
                if (field.type === 'DATETIME') {
                    // Example value 1998-08-02 00:00:00
                    // Here we just omit Date parsing and avoiding Date.toString()
                    // done by driver. MongoBI original format is just fine.
                    return field.string();
                }
                return next();
            },
            // mysql2 v3.x uses this flag by default and sends some connection attributes like:
            // version, app-name. But mongosql which is based on mysql 5.7 is not able to proceed them, resulting in:
            // Error: recv handshake response error: invalid connection attribute at index 0: EOF
            flags: ['-CONNECT_ATTRS'],
            ...mongoBIDriverConfiguration
        };
        this.pool = generic_pool_1.default.createPool({
            create: async () => {
                const conn = (0, mysql2_1.createConnection)(this.config);
                if (conn.on) {
                    conn.on('error', () => {
                        conn.destroy();
                    });
                }
                await conn.promise().connect();
                return conn;
            },
            destroy: async (connection) => connection.promise().end(),
            validate: async (connection) => {
                try {
                    await connection.promise().query({
                        sql: 'SELECT 1',
                    });
                }
                catch (e) {
                    this.databasePoolError(e);
                    return false;
                }
                return true;
            }
        }, {
            min: 0,
            max: config.maxPoolSize ||
                (0, shared_1.getEnv)('dbMaxPoolSize', { dataSource }) ||
                8,
            evictionRunIntervalMillis: 10000,
            softIdleTimeoutMillis: 30000,
            idleTimeoutMillis: 30000,
            testOnBorrow: true,
            acquireTimeoutMillis: 20000
        });
    }
    withConnection(fn) {
        const self = this;
        const connectionPromise = this.pool.acquire();
        let cancelled = false;
        const cancelObj = {};
        const promise = connectionPromise.then(conn => {
            cancelObj.cancel = async () => {
                cancelled = true;
                await self.withConnection(async (processConnection) => {
                    const processRows = await processConnection.promise().query({
                        sql: 'SHOW PROCESSLIST'
                    });
                    await Promise.all(processRows.filter((row) => row.Time >= 599)
                        .map((row) => processConnection.promise().query({
                        sql: `KILL ${row.Id}`
                    })));
                });
            };
            return fn(conn)
                .then(res => this.pool.release(conn).then(() => {
                if (cancelled) {
                    throw new Error('Query cancelled');
                }
                return res;
            }))
                .catch((err) => this.pool.release(conn).then(() => {
                if (cancelled) {
                    throw new Error('Query cancelled');
                }
                throw err;
            }));
        });
        promise.cancel = () => cancelObj.cancel();
        return promise;
    }
    async testConnection() {
        // eslint-disable-next-line no-underscore-dangle
        const conn = await this.pool._factory.create();
        try {
            await conn.promise().query('SELECT 1', []);
        }
        finally {
            // eslint-disable-next-line no-underscore-dangle
            await this.pool._factory.destroy(conn);
        }
    }
    async query(sql, values) {
        return this.withConnection(async (conn) => {
            await this.prepareConnection(conn);
            const [rows] = await conn.promise().query({
                sql,
                values,
            });
            return rows;
        });
    }
    async prepareConnection(conn) {
        await conn.promise().query({
            sql: `SET time_zone = '${this.config.storeTimezone || '+00:00'}'`,
            values: [],
        });
    }
    async stream(query, values, options) {
        // eslint-disable-next-line no-underscore-dangle
        const conn = await this.pool._factory.create();
        try {
            await this.prepareConnection(conn);
            const [rowStream, fields] = await (new Promise((resolve, reject) => {
                const stream = conn.query(query, values).stream(options);
                stream.on('fields', (f) => {
                    resolve([stream, f]);
                });
                stream.on('error', (e) => {
                    reject(e);
                });
            }));
            return {
                rowStream,
                types: fields.map((field) => ({
                    name: field.name,
                    type: this.toGenericType((0, MySQLType_1.getNativeTypeName)(field.columnType)),
                })),
                release: async () => {
                    // eslint-disable-next-line no-underscore-dangle
                    await this.pool._factory.destroy(conn);
                }
            };
        }
        catch (e) {
            // eslint-disable-next-line no-underscore-dangle
            await this.pool._factory.destroy(conn);
            throw e;
        }
    }
    async downloadQueryResults(query, values, options) {
        if ((options || {}).streamImport) {
            return this.stream(query, values, options);
        }
        return this.withConnection(async (conn) => {
            await this.prepareConnection(conn);
            const [rows, fields] = await conn.promise().query(query, values);
            return {
                rows,
                types: fields.map((field) => ({
                    name: field.name,
                    type: this.toGenericType((0, MySQLType_1.getNativeTypeName)(field.columnType)),
                })),
            };
        });
    }
    async release() {
        await this.pool.drain();
        await this.pool.clear();
    }
    informationSchemaQuery() {
        if (this.config.database) {
            return `${super.informationSchemaQuery()} AND columns.table_schema = '${this.config.database}'`;
        }
        else {
            return super.informationSchemaQuery();
        }
    }
    quoteIdentifier(identifier) {
        return `\`${identifier}\``;
    }
    readOnly() {
        // Mongo BI doesn't support table creation
        return true;
    }
}
exports.MongoBIDriver = MongoBIDriver;
//# sourceMappingURL=MongoBIDriver.js.map